/*  player program for ICFP 2002 programming contest  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#define PLAYER_INPUT_BUFFER_SIZE  4096
#define PLAYER_OUTPUT_BUFFER_SIZE 4096

#define PLAYER_TARGET_KIND_NONE                      0
#define PLAYER_TARGET_KIND_POSSIBLE_PACKAGE_LOCATION 1
#define PLAYER_TARGET_KIND_INTACT_HOME_BASE          2
#define PLAYER_TARGET_KIND_DEFINITE_PACKAGE_LOCATION 3
#define PLAYER_TARGET_KIND_OTHER_PACKAGE_DESTINATION 4
#define PLAYER_TARGET_KIND_FIRST_PACKAGE_DESTINATION 5
#define PLAYER_TARGET_KIND_COUNT                     6

static int playerTargetFactors[] = {
   -1,  /*  PLAYER_TARGET_KIND_NONE  */
  150,  /*  PLAYER_TARGET_KIND_POSSIBLE_PACKAGE_LOCATION  */
   75,  /*  PLAYER_TARGET_KIND_INTACT_HOME_BASE  */
   15,  /*  PLAYER_TARGET_KIND_DEFINITE_PACKAGE_LOCATION  */
    3,  /*  PLAYER_TARGET_KIND_OTHER_PACKAGE_DESTINATION  */
    2   /*  PLAYER_TARGET_KIND_FIRST_PACKAGE_DESTINATION  */
};

/*  PlayerTile_t  --  This structure contains information  */
/*  about a single tile of the board.                      */

typedef struct PlayerTile_s {
  struct PlayerTile_s *next;
  char                 mapByte;
  char                 competitorHere;
  char                 targetKind;
  char                 firstMove;
} PlayerTile_t;

/*  PlayerHomeBase_t  --  This structure represents a home base.  */
/*  It will be removed from the list of active home bases when    */
/*  our robot visits the tile and we learn what packages are      */
/*  actually there.                                               */

typedef struct PlayerHomeBase_s {
  struct PlayerHomeBase_s *next;
  struct PlayerHomeBase_s *last;
  PlayerTile_t            *tile;
  char                     targetKind;
} PlayerHomeBase_t;

/*  PlayerRobot_t  --  This structure represents a robot.  There  */
/*  is a balanced tree of robot structures indexed by id.         */

typedef struct PlayerRobot_s {
  struct PlayerRobot_s *next;
  struct PlayerRobot_s *last;
  struct PlayerRobot_s *left;
  struct PlayerRobot_s *right;
  long                  id;
  PlayerTile_t         *tile;
  signed char           balance;
  char                  alive;
} PlayerRobot_t;

/*  PlayerPackage_t  --  This structure represents a package.  There  */
/*  is a balanced tree of package structures indexed by id.           */

typedef struct PlayerPackage_s {
  struct PlayerPackage_s *next;
  struct PlayerPackage_s *last;
  struct PlayerPackage_s *left;
  struct PlayerPackage_s *right;
  long                    id;
  long                    weight;
  PlayerTile_t           *location;
  PlayerTile_t           *destination;
  signed char             balance;
  char                    carried;
} PlayerPackage_t;

static int              playerBestTargetKind;
static int              playerColumnMask;
static long             playerDistance;
static int              playerHeight;
static int              playerInputLength;
static int              playerInputNext;
static PlayerRobot_t   *playerMyRobot;
static PlayerTile_t    *playerOrbit;
static int              playerOutputLength;
static long             playerRemainingCapacity;
static long             playerRemainingMoney;
static int              playerRoundedWidth;
static int              playerRowShift;
static int              playerSocket;
static unsigned char   *playerTileBitMap;
static long             playerTileBitMapLength;
static PlayerTile_t    *playerTiles;
static PlayerPackage_t *playerTopPackage;
static PlayerRobot_t   *playerTopRobot;
static int              playerWidth;

static PlayerHomeBase_t playerActiveHomeBaseHead;
static PlayerRobot_t    playerAliveRobotHead;
static PlayerPackage_t  playerCarriedPackageHead;
static PlayerPackage_t  playerSittingPackageHead;

static long          playerTargetDistances[PLAYER_TARGET_KIND_COUNT];
static PlayerTile_t *playerTargetTiles[PLAYER_TARGET_KIND_COUNT];

static char playerInputBuffer[PLAYER_INPUT_BUFFER_SIZE];
static char playerOutputBuffer[PLAYER_OUTPUT_BUFFER_SIZE];

/*  playerFatal  --  Write an error message, and then exit.  */

static void playerFatal(char *message)
{
  fprintf(stderr, "fatal error: %s\n", message);
  exit(-1);
}

/*  playerOpenSocket  --  Open a socket and connect to the server.  */

static void playerOpenSocket(int argc, char **argv)
{
  struct hostent *host;
  long port;
  char *endp;
  int s;
  struct sockaddr_in server;

  if (argc != 3)
    playerFatal("command line arguments must be host name and port");

  /*  Resolve the host name.  */

  host = gethostbyname(argv[1]);
  if (host == (struct hostent *) 0)
    playerFatal("specified host name could not be resolved");
  if (host->h_addrtype != AF_INET)
    playerFatal("specified host name does not map to Internet address");

  /*  Check the port number.  */

  port = strtol(argv[2], &endp, 10);
  if ((*endp != '\0') || (port < 1) || (port > 65535))
    playerFatal("specified port is invalid");

  /*  Open the socket.  */

  playerSocket = socket(PF_INET, SOCK_STREAM, 0);
  if (playerSocket == -1)
    playerFatal("could not open socket");

  /*  Connect to the server.  */

  memset(&server, 0, sizeof (struct sockaddr_in));
  server.sin_family = AF_INET;
  server.sin_port = htons(port);
  server.sin_addr.s_addr = *((in_addr_t *) host->h_addr);
  s = connect(
    playerSocket,
    (struct sockaddr *) &server,
    sizeof (struct sockaddr_in));
  if (s == -1)
    playerFatal("could not connect to specified server");

  /*  Initialize input and output variables.  */

  playerInputNext = 0;
  playerInputLength = 0;
  playerOutputLength = 0;
}

/*  playerWriteString  --  Write a null-terminated string to the server.  */

static void playerWriteString(char *data)
{
  int n;

  n = strlen(data);
  if (write(playerSocket, data, n) != n)
    playerFatal("could not write to socket");
}

/*  playerFlushBuffer  --  Flush the output buffer to the server.  */

static void playerFlushBuffer(void)
{
  if (playerOutputLength != 0) {
    if (write(playerSocket, playerOutputBuffer, playerOutputLength)
        != playerOutputLength)
    {
      playerFatal("could not write output buffer to socket");
    }
    playerOutputLength = 0;
  }
}

/*  playerWriteStringBuffered  --  Write a null-terminated  */
/*  string to the server, via a buffer.                     */

static void playerWriteStringBuffered(char *data)
{
  int n;

  n = strlen(data);
  if (playerOutputLength + n > PLAYER_OUTPUT_BUFFER_SIZE)
    playerFlushBuffer();
  memcpy(&playerOutputBuffer[playerOutputLength], data, n);
  playerOutputLength += n;
}

/*  playerReadByte  --  Read one byte from the server.  This function   */
/*  exits with good status if the connection to the server was closed.  */

static char playerReadByte(void)
{
  char c;

  if (playerInputNext == playerInputLength) {
    playerInputLength =
      read(playerSocket, playerInputBuffer, PLAYER_INPUT_BUFFER_SIZE);
    if (playerInputLength == 0)
      exit(0);
    if (playerInputLength == -1)
      playerFatal("could not read from socket");
    playerInputNext = 0;
  }
  c = playerInputBuffer[playerInputNext];
  ++playerInputNext;
  return c;
}

/*  playerUnreadByte  --  Arrange to read again the most recently  */
/*  read byte from the server.  This function can only be called   */
/*  once immediately following a call to playerReadByte.           */

static void playerUnreadByte(void)
{
  --playerInputNext;
}

/*  playerReadInteger  --  Read an integer from the server.  */

static long playerReadInteger(void)
{
  long n;
  int negative;
  char c;

  n = 0;
  negative = 0;
  c = playerReadByte();

  /*  Skip over spaces.  */

  while (c == ' ')
    c = playerReadByte();

  /*  For a hyphen, indicate that the number will be negative.  */

  if (c == '-') {
    negative = 1;
    c = playerReadByte();
  }

  /*  The current byte should now be a digit.  */

  if ((c < '0') || (c > '9'))
    playerFatal("invalid integer");

  /*  Loop through the digits.  */

  do {
    n = n * 10 + c - '0';
    c = playerReadByte();
  } while ((c >= '0') && (c <= '9'));

  /*  If the terminator was a newline, arrange for it to be read again.  */

  if (c == '\n') {
    playerUnreadByte();
  } else if (c != ' ') {
    playerFatal("missing space after integer");
  }

  /*  Return the result, negated if a hyphen was given.  */

  if (negative)
    n = - n;
  return n;
}

/*  playerExpectNewline  --  Discard a newline byte from the server.  */

static void playerExpectNewline(void)
{
  char c;

  c = playerReadByte();
  while (c == ' ')
    c = playerReadByte();
  if (c != '\n')
    playerFatal("missing newline");
}

/*  playerInsertRobot  --  Insert a robot structure into a balanced tree.  */

static void playerInsertRobot(
  PlayerRobot_t  *newNode,
  PlayerRobot_t **where,
  int            *outerGrew)
{
  PlayerRobot_t *nodeA;
  PlayerRobot_t *nodeB;
  PlayerRobot_t *nodeC;
  int innerGrew;

  nodeA = *where;
  if (nodeA == (PlayerRobot_t *) 0) {
    /*  The tree was empty, so make the new node the only one.  */

    newNode->left = (PlayerRobot_t *) 0;
    newNode->right = (PlayerRobot_t *) 0;
    newNode->balance = 0;
    *where = newNode;
    *outerGrew = 1;
  } else if (newNode->id < nodeA->id) {
    /*  Add the new node to the left subtree.  */

    playerInsertRobot(newNode, &nodeA->left, &innerGrew);
    if (!innerGrew) {
      *outerGrew = 0;
    } else if (nodeA->balance == 1) {
      /*  The right subtree height was 1 more, and now they are the same.  */

      nodeA->balance = 0;
      *outerGrew = 0;
    } else if (nodeA->balance == 0) {
      /*  The subtree heights were the same, and now the left subtree  */
      /*  height is 1 more.                                            */

      nodeA->balance = -1;
      *outerGrew = 1;
    } else {
      /*  The left subtree height was 1 more, and now it is 2 more,  */
      /*  so we must rebalance.  Examine the left subtree.           */

      nodeB = nodeA->left;
      nodeC = nodeB->right;
      if (nodeB->balance == -1) {
        /*      A              B      */
        /*     / \            / \     */
        /*    B   Z   --->   W   A    */
        /*   / \                / \   */
        /*  W   C              C   Z  */

        nodeA->left = nodeC;
        nodeA->balance = 0;
        nodeB->right = nodeA;
        nodeB->balance = 0;
        *where = nodeB;
      } else {
        /*      A               C       */
        /*     / \            /   \     */
        /*    B   Z   --->   B     A    */
        /*   / \            / \   / \   */
        /*  W   C          W   X Y   Z  */
        /*     / \                      */
        /*    X   Y                     */

        nodeB->right = nodeC->left;
        if (nodeC->balance == 1) {
          nodeB->balance = -1;
        } else {
          nodeB->balance = 0;
        }
        nodeA->left = nodeC->right;
        if (nodeC->balance == -1) {
          nodeA->balance = 1;
        } else {
          nodeA->balance = 0;
        }
        nodeC->left = nodeB;
        nodeC->right = nodeA;
        nodeC->balance = 0;
        *where = nodeC;
      }
      *outerGrew = 0;
    }
  } else {
    /*  Add the new node to the right subtree.  */

    playerInsertRobot(newNode, &nodeA->right, &innerGrew);
    if (!innerGrew) {
      *outerGrew = 0;
    } else if (nodeA->balance == -1) {
      /*  The left subtree height was 1 more, and now they are the same.  */

      nodeA->balance = 0;
      *outerGrew = 0;
    } else if (nodeA->balance == 0) {
      /*  The subtree heights were the same, and now the right subtree  */
      /*  height is 1 more.                                             */

      nodeA->balance = 1;
      *outerGrew = 1;
    } else {
      /*  The right subtree height was 1 more, and now it is 2 more,  */
      /*  so we must rebalance.  Examine the right subtree.           */

      nodeB = nodeA->right;
      nodeC = nodeB->left;
      if (nodeB->balance == 1) {
        /*      A              B      */
        /*     / \            / \     */
        /*    W   B   --->   A   Z    */
        /*       / \        / \       */
        /*      C   Z      W   C      */

        nodeA->right = nodeC;
        nodeA->balance = 0;
        nodeB->left = nodeA;
        nodeB->balance = 0;
        *where = nodeB;
      } else {
        /*      A               C       */
        /*     / \            /   \     */
        /*    W   B   --->   A     B    */
        /*       / \        / \   / \   */
        /*      C   Z      W   X Y   Z  */
        /*     / \                      */
        /*    X   Y                     */

        nodeB->left = nodeC->right;
        if (nodeC->balance == -1) {
          nodeB->balance = 1;
        } else {
          nodeB->balance = 0;
        }
        nodeA->right = nodeC->left;
        if (nodeC->balance == 1) {
          nodeA->balance = -1;
        } else {
          nodeA->balance = 0;
        }
        nodeC->left = nodeA;
        nodeC->right = nodeB;
        nodeC->balance = 0;
        *where = nodeC;
      }
      *outerGrew = 0;
    }
  }
}

/*  playerGetRobot  --  Find or create the robot with the given id.  */

static PlayerRobot_t *playerGetRobot(long id)
{
  PlayerRobot_t *robot;
  int grew;

  /*  See if the robot already exists.  */

  robot = playerTopRobot;
  while (robot != (PlayerRobot_t *) 0) {
    if (id == robot->id)
      return robot;
    if (id < robot->id) {
      robot = robot->left;
    } else {
      robot = robot->right;
    }
  }

  /*  No robot exists with the given id, so create it.  */

  robot = (PlayerRobot_t *) malloc(sizeof (PlayerRobot_t));
  if (robot == (PlayerRobot_t *) 0)
    playerFatal("could not allocate robot structure");
  robot->id = id;
  robot->tile = (PlayerTile_t *) 0;
  robot->alive = 0;
  playerInsertRobot(robot, &playerTopRobot, &grew);
  return robot;
}

/*  playerInsertPackage  --  Insert a package  */
/*  structure into a balanced tree.            */

static void playerInsertPackage(
  PlayerPackage_t  *newNode,
  PlayerPackage_t **where,
  int              *outerGrew)
{
  PlayerPackage_t *nodeA;
  PlayerPackage_t *nodeB;
  PlayerPackage_t *nodeC;
  int innerGrew;

  nodeA = *where;
  if (nodeA == (PlayerPackage_t *) 0) {
    /*  The tree was empty, so make the new node the only one.  */

    newNode->left = (PlayerPackage_t *) 0;
    newNode->right = (PlayerPackage_t *) 0;
    newNode->balance = 0;
    *where = newNode;
    *outerGrew = 1;
  } else if (newNode->id < nodeA->id) {
    /*  Add the new node to the left subtree.  */

    playerInsertPackage(newNode, &nodeA->left, &innerGrew);
    if (!innerGrew) {
      *outerGrew = 0;
    } else if (nodeA->balance == 1) {
      /*  The right subtree height was 1 more, and now they are the same.  */

      nodeA->balance = 0;
      *outerGrew = 0;
    } else if (nodeA->balance == 0) {
      /*  The subtree heights were the same, and now the left subtree  */
      /*  height is 1 more.                                            */

      nodeA->balance = -1;
      *outerGrew = 1;
    } else {
      /*  The left subtree height was 1 more, and now it is 2 more,  */
      /*  so we must rebalance.  Examine the left subtree.           */

      nodeB = nodeA->left;
      nodeC = nodeB->right;
      if (nodeB->balance == -1) {
        /*      A              B      */
        /*     / \            / \     */
        /*    B   Z   --->   W   A    */
        /*   / \                / \   */
        /*  W   C              C   Z  */

        nodeA->left = nodeC;
        nodeA->balance = 0;
        nodeB->right = nodeA;
        nodeB->balance = 0;
        *where = nodeB;
      } else {
        /*      A               C       */
        /*     / \            /   \     */
        /*    B   Z   --->   B     A    */
        /*   / \            / \   / \   */
        /*  W   C          W   X Y   Z  */
        /*     / \                      */
        /*    X   Y                     */

        nodeB->right = nodeC->left;
        if (nodeC->balance == 1) {
          nodeB->balance = -1;
        } else {
          nodeB->balance = 0;
        }
        nodeA->left = nodeC->right;
        if (nodeC->balance == -1) {
          nodeA->balance = 1;
        } else {
          nodeA->balance = 0;
        }
        nodeC->left = nodeB;
        nodeC->right = nodeA;
        nodeC->balance = 0;
        *where = nodeC;
      }
      *outerGrew = 0;
    }
  } else {
    /*  Add the new node to the right subtree.  */

    playerInsertPackage(newNode, &nodeA->right, &innerGrew);
    if (!innerGrew) {
      *outerGrew = 0;
    } else if (nodeA->balance == -1) {
      /*  The left subtree height was 1 more, and now they are the same.  */

      nodeA->balance = 0;
      *outerGrew = 0;
    } else if (nodeA->balance == 0) {
      /*  The subtree heights were the same, and now the right subtree  */
      /*  height is 1 more.                                             */

      nodeA->balance = 1;
      *outerGrew = 1;
    } else {
      /*  The right subtree height was 1 more, and now it is 2 more,  */
      /*  so we must rebalance.  Examine the right subtree.           */

      nodeB = nodeA->right;
      nodeC = nodeB->left;
      if (nodeB->balance == 1) {
        /*      A              B      */
        /*     / \            / \     */
        /*    W   B   --->   A   Z    */
        /*       / \        / \       */
        /*      C   Z      W   C      */

        nodeA->right = nodeC;
        nodeA->balance = 0;
        nodeB->left = nodeA;
        nodeB->balance = 0;
        *where = nodeB;
      } else {
        /*      A               C       */
        /*     / \            /   \     */
        /*    W   B   --->   A     B    */
        /*       / \        / \   / \   */
        /*      C   Z      W   X Y   Z  */
        /*     / \                      */
        /*    X   Y                     */

        nodeB->left = nodeC->right;
        if (nodeC->balance == -1) {
          nodeB->balance = 1;
        } else {
          nodeB->balance = 0;
        }
        nodeA->right = nodeC->left;
        if (nodeC->balance == 1) {
          nodeA->balance = -1;
        } else {
          nodeA->balance = 0;
        }
        nodeC->left = nodeA;
        nodeC->right = nodeB;
        nodeC->balance = 0;
        *where = nodeC;
      }
      *outerGrew = 0;
    }
  }
}

/*  playerGetPackage  --  Find or create the package with the given id.  */

static PlayerPackage_t *playerGetPackage(long id)
{
  PlayerPackage_t *package;
  int grew;

  /*  See if the package already exists.  */

  package = playerTopPackage;
  while (package != (PlayerPackage_t *) 0) {
    if (id == package->id)
      return package;
    if (id < package->id) {
      package = package->left;
    } else {
      package = package->right;
    }
  }

  /*  No package exists with the given id, so create it.  */

  package = (PlayerPackage_t *) malloc(sizeof (PlayerPackage_t));
  if (package == (PlayerPackage_t *) 0)
    playerFatal("could not allocate package structure");
  package->id = id;
  package->weight = 0;
  package->location = (PlayerTile_t *) 0;
  package->destination = (PlayerTile_t *) 0;
  package->carried = 0;
  playerInsertPackage(package, &playerTopPackage, &grew);
  return package;
}

/*  playerReadUpdate  --  Read an update from the server.  */

static void playerReadUpdate(void)
{
  PlayerHomeBase_t *homeBase;
  PlayerPackage_t *package;
  PlayerRobot_t *robot;
  long x;
  long y;
  char c;

  /*  As part of the update, we will rebuild the  */
  /*  list of robots that are still alive.        */

  playerAliveRobotHead.next = &playerAliveRobotHead;
  playerAliveRobotHead.last = &playerAliveRobotHead;

  /*  Process bytes from the server until a newline is found.  */

  robot = (PlayerRobot_t *) 0;
  homeBase = (PlayerHomeBase_t *) 0;
  while (1) {
    c = playerReadByte();
    switch (c) {
    case '\n':
      break;

    case ' ':
      continue;

    case '#':
      robot = playerGetRobot(playerReadInteger());
      if (!robot->alive) {
        /*  Add it to the list of robots that are alive.  */

        robot->alive = 1;
        robot->next = &playerAliveRobotHead;
        robot->last = playerAliveRobotHead.last;
        playerAliveRobotHead.last->next = robot;
        playerAliveRobotHead.last = robot;
      }
      continue;

    case 'D':
      if (robot == (PlayerRobot_t *) 0)
        playerFatal("D update before robot specified");
      if (robot->tile == (PlayerTile_t *) 0)
        playerFatal("D update for robot with unknown location");
      package = playerGetPackage(playerReadInteger());
      if (package->location != (PlayerTile_t *) 0)
        playerFatal("D update for package on ground");
      if (!package->carried) {
        if (robot == playerMyRobot)
          playerFatal("D update for my robot, but not carried");
      } else {
        if (robot != playerMyRobot)
          playerFatal("D update for competitor, but carried");

        /*  Remove this package from the carried list.  */

        playerRemainingCapacity += package->weight;
        package->carried = 0;
        package->last->next = package->next;
        package->next->last = package->last;
      }
      if (robot->tile != package->destination) {
        /*  The package has been dropped somewhere other than  */
        /*  its destination.  Add it to the list of packages   */
        /*  sitting on the ground.                             */

        package->location = robot->tile;
        package->next = &playerSittingPackageHead;
        package->last = playerSittingPackageHead.last;
        playerSittingPackageHead.last->next = package;
        playerSittingPackageHead.last = package;
      }
      continue;

    case 'E':
      if (robot == (PlayerRobot_t *) 0)
        playerFatal("E update before robot specified");
      if (robot->tile == (PlayerTile_t *) 0)
        playerFatal("E update for robot with unknown location");
      ++robot->tile;
      continue;

    case 'N':
      if (robot == (PlayerRobot_t *) 0)
        playerFatal("N update before robot specified");
      if (robot->tile == (PlayerTile_t *) 0)
        playerFatal("N update for robot with unknown location");
      robot->tile += playerRoundedWidth;
      continue;

    case 'P':
      if (robot == (PlayerRobot_t *) 0)
        playerFatal("P update before robot specified");
      if (robot->tile == (PlayerTile_t *) 0)
        playerFatal("P update for robot with unknown location");
      package = playerGetPackage(playerReadInteger());
      if (package->carried)
        playerFatal("P update for carried package");
      if (package->location != (PlayerTile_t *) 0) {
        /*  Remove this package from the sitting list.  */

        package->location = (PlayerTile_t *) 0;
        package->last->next = package->next;
        package->next->last = package->last;
      }
      if (robot == playerMyRobot) {
        if (package->destination == (PlayerTile_t *) 0)
          playerFatal("P update for my robot, but unknown destination");

        /*  Add this package to the carried list.  */

        playerRemainingCapacity -= package->weight;
        package->carried = 1;
        package->next = &playerCarriedPackageHead;
        package->last = playerCarriedPackageHead.last;
        playerCarriedPackageHead.last->next = package;
        playerCarriedPackageHead.last = package;
      } else if ((robot->tile->mapByte == '@') &&
                 ((homeBase == (PlayerHomeBase_t *) 0) ||
                  (homeBase->tile != robot->tile)))
      {
        /*  A competitor has picked up a package from a home base.  */
        /*  Update the home base if it is still active, to show     */
        /*  that it is no longer intact.                            */

        homeBase = playerActiveHomeBaseHead.next;
        while (1) {
          if (homeBase == &playerActiveHomeBaseHead) {
            homeBase = (PlayerHomeBase_t *) 0;
            break;
          }
          if (homeBase->tile == robot->tile) {
            homeBase->targetKind =
              PLAYER_TARGET_KIND_POSSIBLE_PACKAGE_LOCATION;
            break;
          }
          homeBase = homeBase->next;
        }
      }
      continue;

    case 'S':
      if (robot == (PlayerRobot_t *) 0)
        playerFatal("S update before robot specified");
      if (robot->tile == (PlayerTile_t *) 0)
        playerFatal("S update for robot with unknown location");
      robot->tile -= playerRoundedWidth;
      continue;

    case 'W':
      if (robot == (PlayerRobot_t *) 0)
        playerFatal("W update before robot specified");
      if (robot->tile == (PlayerTile_t *) 0)
        playerFatal("W update for robot with unknown location");
      --robot->tile;
      continue;

    case 'X':
      if (robot == (PlayerRobot_t *) 0)
        playerFatal("X update before robot specified");
      x = playerReadInteger();
      if ((x < 1) || (x > playerWidth))
        playerFatal("invalid X coordinate");
      c = playerReadByte();
      while (c == ' ')
        c = playerReadByte();
      if (c != 'Y')
        playerFatal("X update without following Y update");
      y = playerReadInteger();
      if ((y < 1) || (y > playerHeight))
        playerFatal("invalid Y coordinate");
      robot->tile = &playerTiles[((y - 1) << playerRowShift) + x - 1];
      continue;

    default:
      playerFatal("unknown update from server");
    }
    break;
  }
}

/*  playerGetConfiguration  --  Get the configuration from the server.  */

static void playerGetConfiguration(void)
{
  PlayerHomeBase_t *homeBase;
  PlayerTile_t *tile;
  long n;
  int i;
  int j;

  /*  Initialize configuration variables.  */

  playerActiveHomeBaseHead.next = &playerActiveHomeBaseHead;
  playerActiveHomeBaseHead.last = &playerActiveHomeBaseHead;
  playerCarriedPackageHead.next = &playerCarriedPackageHead;
  playerCarriedPackageHead.last = &playerCarriedPackageHead;
  playerSittingPackageHead.next = &playerSittingPackageHead;
  playerSittingPackageHead.last = &playerSittingPackageHead;
  playerTopPackage = (PlayerPackage_t *) 0;
  playerTopRobot = (PlayerRobot_t *) 0;

  /*  Ask the server for the configuration.  */

  playerWriteString("Player\n");

  /*  Read the width.  */

  n = playerReadInteger();
  if ((n < 1) || (n > 1000))
    playerFatal("invalid width");
  playerWidth = n;

  /*  Read the height.  */

  n = playerReadInteger();
  if ((n < 1) || (n > 1000))
    playerFatal("invalid height");
  playerHeight = n;
  playerExpectNewline();

  /*  Round the width up to a power of two.  */

  playerRowShift = 0;
  playerRoundedWidth = 1;
  while (playerRoundedWidth < playerWidth) {
    ++playerRowShift;
    playerRoundedWidth <<= 1;
  }
  playerColumnMask = playerRoundedWidth - 1;

  /*  Allocate the array of tiles.  */

  n <<= playerRowShift;
  playerTiles = (PlayerTile_t *) malloc(n * sizeof (PlayerTile_t));
  if (playerTiles == (PlayerTile_t *) 0)
    playerFatal("could not allocate array of tiles");

  /*  Allocate the tile bit map.  */

  playerTileBitMapLength = (n + 7) >> 3;
  playerTileBitMap = (unsigned char *) malloc(playerTileBitMapLength);
  if (playerTileBitMap == (unsigned char *) 0)
    playerFatal("could not allocate tile bit map");

  /*  Read the board from the server.  */

  for (j = 0; j < playerHeight; ++j) {
    tile = &playerTiles[((long) j) << playerRowShift];
    for (i = 0; i < playerWidth; ++i) {
      /*  Read the next byte of the board map from the server.  */

      tile->mapByte = playerReadByte();
      tile->competitorHere = 0;
      tile->targetKind = PLAYER_TARGET_KIND_NONE;
      if (tile->mapByte == '@') {
        /*  Allocate a home base structure and add it to the active list.  */

        homeBase = (PlayerHomeBase_t *) malloc(sizeof (PlayerHomeBase_t));
        if (homeBase == (PlayerHomeBase_t *) 0)
          playerFatal("could not allocate home base structure");
        homeBase->tile = tile;
        homeBase->targetKind = PLAYER_TARGET_KIND_INTACT_HOME_BASE;
        homeBase->next = &playerActiveHomeBaseHead;
        homeBase->last = playerActiveHomeBaseHead.last;
        playerActiveHomeBaseHead.last->next = homeBase;
        playerActiveHomeBaseHead.last = homeBase;
      }
      ++tile;
    }
    playerExpectNewline();
  }

  /*  Read our robot id, and allocate a robot structure for it.  */

  playerMyRobot = playerGetRobot(playerReadInteger());

  /*  Read the weight limit and the amount of money.  */

  playerRemainingCapacity = playerReadInteger();
  playerRemainingMoney = playerReadInteger();
  playerExpectNewline();

  /*  Read the first update, which should give the location of each robot.  */

  playerReadUpdate();
}

/*  playerReadPackageList  --  Read the list of packages  */
/*  on the ground at our current robot location.          */

static void playerReadPackageList(void)
{
  PlayerPackage_t *package;
  long x;
  long y;
  char c;

  if (playerMyRobot->tile == (PlayerTile_t *) 0)
    playerFatal("my robot has unknown location");
  while (1) {
    /*  The end of the package list is indicated by a newline.  */

    c = playerReadByte();
    while (c == ' ')
      c = playerReadByte();
    if (c == '\n')
      break;
    playerUnreadByte();

    /*  Get the package for this id.  */

    package = playerGetPackage(playerReadInteger());
    if (package->carried)
      playerFatal("package on ground, but carried");
    if (package->location == (PlayerTile_t *) 0) {
      /*  Add it to the list of packages sitting on the ground.  */

      package->next = &playerSittingPackageHead;
      package->last = playerSittingPackageHead.last;
      playerSittingPackageHead.last->next = package;
      playerSittingPackageHead.last = package;
    }
    package->location = playerMyRobot->tile;

    /*  Get the horizontal position of the destination.  */

    x = playerReadInteger();
    if ((x < 1) || (x > playerWidth))
      playerFatal("invalid horizontal position for package destination");

    /*  Get the vertical position of the destination.  */

    y = playerReadInteger();
    if ((y < 1) || (y > playerHeight))
      playerFatal("invalid vertical position for package destination");
    package->destination = &playerTiles[((y - 1) << playerRowShift) + x - 1];

    /*  Get the package weight.  */

    package->weight = playerReadInteger();
  }
}

/*  playerConsiderMove  --  Consider moving to the given tile.  */

static void playerConsiderMove(PlayerTile_t *tile, char firstMove)
{
  if ((tile->mapByte != '~') &&
      (tile->mapByte != '#') &&
      ((tile->targetKind != PLAYER_TARGET_KIND_NONE) || !tile->competitorHere))
  {
    /*  Add this tile to the orbit for the current distance.  */

    tile->firstMove = firstMove;
    tile->next = playerOrbit;
    playerOrbit = tile;

    /*  Record this tile if it is the first of its target kind.  */

    if (playerTargetTiles[tile->targetKind] == (PlayerTile_t *) 0) {
      playerTargetTiles[tile->targetKind] = tile;
      playerTargetDistances[tile->targetKind] = playerDistance;
    }
  }
}

/*  playerIssueMoveCommand  --  Issue a move command to the server.  */

static void playerIssueMoveCommand(void)
{
  PlayerTile_t *finalTarget;
  PlayerTile_t *tile;
  long finalScore;
  long score;
  long m;
  long n;
  long n2;
  int b;
  int i;
  int j;
  int k;
  char temp[16];

  memset(playerTileBitMap, 0, playerTileBitMapLength);
  for (k = 0; k < PLAYER_TARGET_KIND_COUNT; ++k)
    playerTargetTiles[k] = (PlayerTile_t *) 0;
  n = playerMyRobot->tile - playerTiles;
  playerTileBitMap[n >> 3] = 1 << (n & 7);

  /*  Build the orbit for distance 1.  */

  playerDistance = 1;
  playerOrbit = (PlayerTile_t *) 0;
  j = n >> playerRowShift;
  i = n & playerColumnMask;
  if (j != playerHeight - 1) {
    /*  Consider moving north.  */

    n2 = n + playerRoundedWidth;
    playerTileBitMap[n2 >> 3] |= 1 << (n2 & 7);
    playerConsiderMove(&playerTiles[n2], 'N');
  }
  if (i != playerWidth - 1) {
    /*  Consider moving east.  */

    n2 = n + 1;
    playerTileBitMap[n2 >> 3] |= 1 << (n2 & 7);
    playerConsiderMove(&playerTiles[n2], 'E');
  }
  if (j != 0) {
    /*  Consider moving south.  */

    n2 = n - playerRoundedWidth;
    playerTileBitMap[n2 >> 3] |= 1 << (n2 & 7);
    playerConsiderMove(&playerTiles[n2], 'S');
  }
  if (i != 0) {
    /*  Consider moving west.  */

    n2 = n - 1;
    playerTileBitMap[n2 >> 3] |= 1 << (n2 & 7);
    playerConsiderMove(&playerTiles[n2], 'W');
  }

  /*  Keep building orbits until our best target is reached,  */
  /*  or no further motion is possible.                       */

  while ((playerOrbit != (PlayerTile_t *) 0) &&
         (playerTargetTiles[playerBestTargetKind] == (PlayerTile_t *) 0))
  {
    tile = playerOrbit;
    ++playerDistance;
    playerOrbit = (PlayerTile_t *) 0;
    do {
      if (!tile->competitorHere) {
        n = tile - playerTiles;
        j = n >> playerRowShift;
        i = n & playerColumnMask;
        if (j != playerHeight - 1) {
          /*  Consider moving north.  */

          n2 = n + playerRoundedWidth;
          m = n2 >> 3;
          b = 1 << (n2 & 7);
          if (!(playerTileBitMap[m] & b)) {
            playerTileBitMap[m] |= b;
            playerConsiderMove(&playerTiles[n2], tile->firstMove);
          }
        }
        if (i != playerWidth - 1) {
          /*  Consider moving east.  */

          n2 = n + 1;
          m = n2 >> 3;
          b = 1 << (n2 & 7);
          if (!(playerTileBitMap[m] & b)) {
            playerTileBitMap[m] |= b;
            playerConsiderMove(&playerTiles[n2], tile->firstMove);
          }
        }
        if (j != 0) {
          /*  Consider moving south.  */

          n2 = n - playerRoundedWidth;
          m = n2 >> 3;
          b = 1 << (n2 & 7);
          if (!(playerTileBitMap[m] & b)) {
            playerTileBitMap[m] |= b;
            playerConsiderMove(&playerTiles[n2], tile->firstMove);
          }
        }
        if (i != 0) {
          /*  Consider moving west.  */

          n2 = n - 1;
          m = n2 >> 3;
          b = 1 << (n2 & 7);
          if (!(playerTileBitMap[m] & b)) {
            playerTileBitMap[m] |= b;
            playerConsiderMove(&playerTiles[n2], tile->firstMove);
          }
        }
      }
      tile = tile->next;
    } while ((tile != (PlayerTile_t *) 0) &&
             (playerTargetTiles[playerBestTargetKind] == (PlayerTile_t *) 0));
  }

  /*  Choose the final target.  */

  finalTarget = (PlayerTile_t *) 0;
  for (k = PLAYER_TARGET_KIND_POSSIBLE_PACKAGE_LOCATION;
       k <= PLAYER_TARGET_KIND_FIRST_PACKAGE_DESTINATION;
       ++k)
  {
    tile = playerTargetTiles[k];
    if (tile != (PlayerTile_t *) 0) {
      score = playerTargetDistances[k] * playerTargetFactors[k];
      if ((finalTarget == (PlayerTile_t *) 0) || (score < finalScore)) {
        finalTarget = tile;
        finalScore = score;
      }
    }
  }
  if (finalTarget == (PlayerTile_t *) 0) {
    /*  No targets are reachable, perhaps because they are currently  */
    /*  blocked by other robots.  Tell our robot not to do anything.  */

    playerRemainingMoney -= 1;
    playerWriteString("1 Pick\n");
    return;
  }

  /*  Ask to make the first move toward the final target.  */

  playerRemainingMoney -= 1;
  sprintf(temp, "1 Move %c\n", finalTarget->firstMove);
  playerWriteString(temp);
}

/*  playerIssuePickCommand  --  Issue a pick command to the server.  */

static void playerIssuePickCommand(void)
{
  PlayerPackage_t *package;
  long capacity;
  char temp[16];

  /*  Ask to pick up whatever packages our robot can carry.  */

  playerRemainingMoney -= 1;
  playerWriteStringBuffered("1 Pick");
  capacity = playerRemainingCapacity;
  package = playerSittingPackageHead.next;
  while (package != &playerSittingPackageHead) {
    if ((package->location == playerMyRobot->tile) &&
        (package->weight <= capacity))
    {
      capacity -= package->weight;
      sprintf(temp, " %ld", package->id);
      playerWriteStringBuffered(temp);
    }
    package = package->next;
  }
  playerWriteStringBuffered("\n");
  playerFlushBuffer();
}

/*  playerIssueDropCommand  --  Issue a drop command to the server.  */

static void playerIssueDropCommand(void)
{
  PlayerPackage_t *package;
  char temp[16];

  /*  Ask to drop all packages destined for the current robot location.  */

  playerRemainingMoney -= 1;
  playerWriteStringBuffered("1 Drop");
  package = playerCarriedPackageHead.next;
  while (package != &playerCarriedPackageHead) {
    if (package->destination == playerMyRobot->tile) {
      sprintf(temp, " %ld", package->id);
      playerWriteStringBuffered(temp);
    }
    package = package->next;
  }
  playerWriteStringBuffered("\n");
  playerFlushBuffer();
}

/*  playerIssueCommand  --  Choose a command and issue it.  On entry, the  */
/*  competitorHere and targetKind fields of all tile structures are 0.     */
/*  On exit, the same is true, and in addition, all robots are marked as   */
/*  no longer alive, since playerReadUpdate will rebuild the list of       */
/*  robots that are alive.                                                 */

static void playerIssueCommand(void)
{
  PlayerHomeBase_t *homeBase;
  PlayerPackage_t *package;
  PlayerRobot_t *robot;
  PlayerTile_t *tile;

  /*  Mark tiles where competitors are located.  */

  robot = playerAliveRobotHead.next;
  while (robot != &playerAliveRobotHead) {
    if ((robot != playerMyRobot) && (robot->tile != (PlayerTile_t *) 0))
      robot->tile->competitorHere = 1;
    robot = robot->next;
  }

  /*  Mark tiles for active home bases.  If our robot is currently   */
  /*  at a home base, then mark that home base as no longer active.  */

  playerBestTargetKind = PLAYER_TARGET_KIND_NONE;
  homeBase = playerActiveHomeBaseHead.next;
  while (homeBase != &playerActiveHomeBaseHead) {
    if (homeBase->tile != playerMyRobot->tile) {
      if (playerRemainingCapacity != 0) {
        homeBase->tile->targetKind = homeBase->targetKind;
        if (homeBase->targetKind > playerBestTargetKind)
          playerBestTargetKind = homeBase->targetKind;
      }
    } else {
      /*  We now know all the packages here, so remove  */
      /*  this home base from the active list.          */

      homeBase->targetKind = PLAYER_TARGET_KIND_NONE;
      homeBase->last->next = homeBase->next;
      homeBase->next->last = homeBase->last;
    }
    homeBase = homeBase->next;
  }

  /*  Mark tiles for packages on the ground.  */

  package = playerSittingPackageHead.next;
  while (package != &playerSittingPackageHead) {
    if (package->destination == (PlayerTile_t *) 0) {
      tile = package->location;
      if (tile == playerMyRobot->tile) {
        /*  This package must have been dropped at its destination,  */
        /*  since otherwise it would have appeared in the list from  */
        /*  the server.  Remove it from the sitting list.            */

        package->location = (PlayerTile_t *) 0;
        package->last->next = package->next;
        package->next->last = package->last;
      } else if ((playerRemainingCapacity != 0) &&
                 (tile->targetKind == PLAYER_TARGET_KIND_NONE))
      {
        tile->targetKind = PLAYER_TARGET_KIND_POSSIBLE_PACKAGE_LOCATION;
        if (playerBestTargetKind == PLAYER_TARGET_KIND_NONE)
          playerBestTargetKind = PLAYER_TARGET_KIND_POSSIBLE_PACKAGE_LOCATION;
      }
    } else if (package->weight <= playerRemainingCapacity) {
      package->location->targetKind =
        PLAYER_TARGET_KIND_DEFINITE_PACKAGE_LOCATION;
      playerBestTargetKind = PLAYER_TARGET_KIND_DEFINITE_PACKAGE_LOCATION;
    }
    package = package->next;
  }

  /*  Mark tiles for destinations of packages our robot is carrying.  */

  package = playerCarriedPackageHead.next;
  if (package != &playerCarriedPackageHead) {
    playerBestTargetKind = PLAYER_TARGET_KIND_FIRST_PACKAGE_DESTINATION;
    package->destination->targetKind =
      PLAYER_TARGET_KIND_FIRST_PACKAGE_DESTINATION;
    package = package->next;
    while (package != &playerCarriedPackageHead) {
      tile = package->destination;
      if (tile->targetKind < PLAYER_TARGET_KIND_OTHER_PACKAGE_DESTINATION)
        tile->targetKind = PLAYER_TARGET_KIND_OTHER_PACKAGE_DESTINATION;
      package = package->next;
    }
  }

  /*  Issue an appropriate command based on our current robot location.  */

  switch (playerMyRobot->tile->targetKind) {
  case PLAYER_TARGET_KIND_DEFINITE_PACKAGE_LOCATION:
    playerIssuePickCommand();
    break;

  case PLAYER_TARGET_KIND_OTHER_PACKAGE_DESTINATION:
  case PLAYER_TARGET_KIND_FIRST_PACKAGE_DESTINATION:
    playerIssueDropCommand();
    break;

  default:
    if (playerBestTargetKind != PLAYER_TARGET_KIND_NONE) {
      playerIssueMoveCommand();
    } else {
      /*  There are no more targets at all.  The game should be  */
      /*  over soon.  Tell our robot not to do anything.         */

      playerRemainingMoney -= 1;
      playerWriteString("1 Pick\n");
    }
  }

  /*  Unmark tiles for destinations of packages our robot is carrying.  */

  package = playerCarriedPackageHead.next;
  while (package != &playerCarriedPackageHead) {
    package->destination->targetKind = PLAYER_TARGET_KIND_NONE;
    package = package->next;
  }

  /*  Unmark tiles for packages on the ground.  */

  package = playerSittingPackageHead.next;
  while (package != &playerSittingPackageHead) {
    package->location->targetKind = PLAYER_TARGET_KIND_NONE;
    package = package->next;
  }

  /*  Unmark tiles for active home bases.  */

  homeBase = playerActiveHomeBaseHead.next;
  while (homeBase != &playerActiveHomeBaseHead) {
    homeBase->tile->targetKind = PLAYER_TARGET_KIND_NONE;
    homeBase = homeBase->next;
  }

  /*  Unmark tiles where competitors are located.  */
  /*  Also, mark all robots as no longer alive.    */

  robot = playerAliveRobotHead.next;
  while (robot != &playerAliveRobotHead) {
    if (robot->tile != (PlayerTile_t *) 0)
      robot->tile->competitorHere = 0;
    robot->alive = 0;
    robot = robot->next;
  }
}

int main(int argc, char **argv)
{
  playerOpenSocket(argc, argv);
  playerGetConfiguration();
  while (playerMyRobot->alive) {
    playerReadPackageList();
    playerIssueCommand();
    playerReadUpdate();
  }
  return 0;
}
